(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Arith.

Require Import list_fall_exst.
Require Import predicate.
Require Import closure.
Require Import maps.
Require Import carrier. 
Require Import finite.

Require Import labels.
Require Import formula.
Require Import words.
Require Import constraints.
Require Import pme.
Require Import css.

Require Import branch.
Require Import bbi_expansion.

Proposition bbi_expand_1_id : forall CSS1 CSS2 CSS3 CSS4, CSS_id CSS1 CSS2 -> CSS_id CSS3 CSS4 -> CSS_id (bbi_expand_1 CSS1 CSS3) (bbi_expand_1 CSS2 CSS4).
Proof. intros [] [] [] [] [] []; simpl; split; simpl in * |- *; auto. Qed.

Hint Resolve bbi_expand_1_id.
Hint Unfold bbi_expand_1.

Proposition bbi_expand_id css1 css2 : CSS_id css1 css2 -> forall l1 l2, listbranch_eq CSS_id l1 l2 -> listbranch_eq CSS_id (bbi_expand css1 l1) (bbi_expand css2 l2).
Proof.
  intros G; induction l1 as [ | x l1 ]; intros [ | y l2 ] H; auto.
  simpl in H |- *. destruct H.
  split; auto.
Qed.

Proposition bbi_expand_increase : forall css lCSS cssi, carrier (bbi_expand css lCSS) cssi -> CSS_leq css cssi.
Proof.
  intros []; induction lCSS as [ | [] lCSS ]; intros [] H; simpl in H; try contradiction H.
  destruct H as [ H | ]. unfold bbi_expand_1 in H. 
  rewrite <- H; simpl; auto.
  apply IHlCSS; auto.
Qed.

Proposition bbi_expand_app CSS l1 l2 : bbi_expand CSS (l1++l2) = (bbi_expand CSS l1)++(bbi_expand CSS l2).
Proof. apply map_app. Qed.

Hint Rewrite bbi_expand_app.

Proposition bbi_expand_comp X1 C1 X2 C2 lcss : listbranch_eq CSS_id (bbi_expand (X1,C1) (bbi_expand (X2,C2) lcss)) 
                                                                    (bbi_expand (X1 cup X2, C1 cup C2) lcss).
Proof. induction lcss as [ | [] ]; simpl; unfold bbi_expand_1; auto. Qed.
   
Proposition bbi_tab_constraints_app l1 l2 : bbi_tab_constraints (l1++l2) ~ bbi_tab_constraints l1 cup bbi_tab_constraints l2.
Proof.
  induction l1 as [ | [] ]; simpl; auto.
  apply id_transitive with (2 := bin_union_assoc _ _ _ _).
  apply bin_union_congr; auto.
Qed.

Hint Resolve bbi_expand_comp.
Hint Immediate bbi_tab_constraints_app.

Proposition bbi_expand_inc css lcss : bbi_tab_constraints lcss inc bbi_tab_constraints (bbi_expand css lcss).
Proof. destruct css; induction lcss as [ | [] ]; simpl; auto. Qed.

Hint Resolve bbi_expand_inc.

Proposition bbi_tab_constraints_id : forall l1 l2, listbranch_eq CSS_id l1 l2 -> bbi_tab_constraints l1 ~ bbi_tab_constraints l2.
Proof.
  induction l1 as [ | [] ]; destruct l2 as [ | [] ]; auto; intros H; try contradiction H.
  simpl in H |- *. destruct H as [ [] H1 ]. generalize (IHl1 _ H1).
  intro; auto.
Qed.

Proposition bbi_tab_constraints_expand_inc X C lcss : lcss <> nil -> C inc bbi_tab_constraints (bbi_expand (X,C) lcss).
Proof. 
  intros H1; induction lcss as [ | [] ]; try (contradict H1; auto; fail).
  simpl; apply inc_transitive with (2 := bin_union_inc_r1 _ _ _); auto.
Qed.

Section listbranch_eq_expand_CSS_inc.

  Implicit Types A : cst -> Prop.

  Let pme_lemma A B C D : pme A inc pme B -> pme C inc pme D -> pme (A cup C) inc pme (B cup D).
  Proof.
    intros H1 H2.
    apply pme_min; auto; pred_solve;
    apply inc_transitive with (1 := pme_inc _);
    [ apply inc_transitive with (1 := H1) 
    | apply inc_transitive with (1 := H2)
    ]; auto.
  Qed.

  Let alphabet_lemma_2 A B C D : alphabet A ~ alphabet B -> alphabet C ~ alphabet D -> alphabet (A cup C) ~ alphabet (B cup D).
  Proof. intros; do 2 (apply id_transitive with (1 := @alphabet_cup _ _ _); apply id_symetric); auto. Qed.
 
  Proposition listbranch_eq_expand_CSS_inc r p css css' : In r bbi_exp_rules -> CSS_inc css css' -> listbranch_eq CSS_inc (bbi_expand css (bbi_inst r p)) (bbi_expand css' (bbi_inst r p)).
  Proof.
    destruct css; destruct css'; intros Hr ( ? & ? & ? ); unfold fst, snd in * |- *.
    unfold bbi_exp_rules in Hr; In_auto; symmetry in Hr; 
    subst; destruct p as ((((((A,B),a),b),m),x),y);
     simpl; unfold bbi_expand_1; repeat (simpl; auto; split; auto).
  Qed.

End listbranch_eq_expand_CSS_inc.

Proposition bbi_tab_constraints_alphabet l1 l2 : alphabet (bbi_tab_constraints (l1++l2)) ~ alphabet (bbi_tab_constraints l1) cup alphabet (bbi_tab_constraints l2).
Proof. apply id_transitive with (2 := @alphabet_cup _ _ _); auto. Qed.

Hint Immediate bbi_tab_constraints_alphabet.

Proposition bbi_tab_constraints_subst s A B ll : (forall x, carrier (s x) cap A inc B) -> alphabet (bbi_tab_constraints (map (subst_CSS s) ll)) cap A inc B.
Proof.
  intros Hs. induction ll as [ | (X,C) ll ]; simpl;
  intros x Hx; apply bin_inter_sg in Hx.
 
  destruct Hx as [ Hx _ ]; apply alphabet_empty in Hx; contradiction Hx.

  destruct Hx as [ H1 H2 ].
  apply alphabet_cup, bin_union_sg in H1.
  destruct H1 as [ H1 | H1 ].

  apply subst_dimage_alphabet in H1.
  apply subst_alphabet_prop in H1.
  destruct H1 as ( a & H3 & H4 ).

  apply (Hs a), bin_inter_sg'; auto.
  
  apply IHll, bin_inter_sg'; auto.
Qed.

Hint Resolve bbi_tab_constraints_subst.
    
Proposition bbi_fireable_inc X X' C D r p : X ~ X' -> pme C inc pme D -> alphabet C ~ alphabet D -> In r bbi_exp_rules-> bbi_fireable r p (X,C) -> bbi_fireable r p (X',D).
Proof.
  intros [H1 H1'] H2 [ H3 H4 ] Hr Hp.
  unfold bbi_exp_rules in Hr; In_auto; symmetry in Hr; 
  subst; destruct p as ((((((A,B),a),b),m),x),y); 
  simpl in Hp |- *; auto;
  red in Hp |- *;
  unfold Included in H1, H2, H3, H4; intuition.
Qed.

Proposition bbi_fireable_equiv X C X' C' r p : CSS_equiv (X,C) (X',C') -> In r bbi_exp_rules -> bbi_fireable r p (X,C) -> bbi_fireable r p (X',C').
Proof.
  intros [ ? [ ? ? ] ]; apply bbi_fireable_inc; auto.
  do 2 apply id_transitive with (1 := alphabet_pme _), id_symetric; auto.
Qed.

Proposition bbi_expansion_not_empty css lcss : bbi_expansion css lcss -> lcss <> nil.
Proof.
  destruct css as (X,C); induction 1 as [ r p Hr Hp ]. 
  unfold bbi_exp_rules in Hr; In_auto; symmetry in Hr; 
  subst; destruct p as ((((((A,B),a),b),m),x),y); simpl in Hp |- *;
  discriminate.
Qed.

Proposition bbi_expansion_increase' css lcss : bbi_expansion css lcss -> forall cssi, carrier lcss cssi -> CSS_leq css cssi.
Proof. induction 1; auto; apply bbi_expand_increase. Qed.

Proposition bbi_expansion_increase X C lcss : bbi_expansion (X,C) lcss -> C inc bbi_tab_constraints lcss.
Proof.
  intros H.
  destruct lcss as [ | (X',C') lcss ].
  apply bbi_expansion_not_empty in H; contradict H; auto.
  generalize (bbi_expansion_increase' _ _ H (X',C') (carrier_cons_hd _ _ _)); intros [ H1 H2 ].
  simpl; pred_solve.
Qed.

Hint Resolve bbi_expansion_increase.

Proposition bbi_expansion_finite css lcss : finite_pair css -> bbi_expansion css lcss -> carrier lcss inc finite_pair.
Proof.
  destruct css as (X,C); intros [ H1 H2 ] [ r p Hr Hp ].
  unfold bbi_exp_rules in Hr; In_auto; symmetry in Hr; 
  subst; destruct p as ((((((A,B),a),b),m),x),y); simpl in Hp |- *; unfold bbi_expand_1; auto.
Qed.

Section bbi_expansion_CSS.

  Let bbi_exp_lemma_1 (X : _ -> Prop) C s f m : X (s,(f,m)) -> CSS (X,C) -> pme C (m~~m).
  Proof. intros H1 H2; apply H2 in H1; auto. Qed.

  Proposition bbi_expansion_CSS css lcss : CSS css -> bbi_expansion css lcss -> carrier lcss inc CSS.
  Proof.
    destruct css as (X,C); intros HCSS [ r p Hr Hp ].
    unfold bbi_exp_rules in Hr; In_auto; symmetry in Hr; 
    subst; destruct p as ((((((A,B),a),b),m),x),y); 
    simpl in Hp; try red in Hp; simpl; unfold bbi_expand_1;
    repeat match goal with [ Hp : _ /\ _ |- _ ] => destruct Hp as [ ? Hp ] end;
    match goal with [ H: X _ |- _ ] => generalize (HCSS _ _ _ H); intros H10 end;
    try match goal with [ H: pme _ (_ o _ ~~ _) |- _ ] => apply pme_help_lemma_9 in H; destruct H as [ H [] ] end;
    repeat apply carrier_cons_2; auto. 
    
    repeat (apply CSS_cup_l; auto); apply CSS_sg_l.
    apply pme_help_lemma_2 with (y := l b) (z := m), pme_inc; simpl; pred_solve.
    apply pme_help_lemma_1 with (x := l a) (z := m), pme_inc; simpl; pred_solve.
    
    repeat (apply CSS_cup_l; auto); apply CSS_sg_l.
    apply pme_help_lemma_2 with (y := m) (z := l b), pme_inc; simpl; pred_solve.
    apply pme_help_lemma_4 with (x := l a) (y := m), pme_inc; simpl; pred_solve.
  Qed.

End bbi_expansion_CSS.
 
Proposition bbi_expansion_finite_CSS css lcss : finite_CSS css -> bbi_expansion css lcss -> carrier lcss inc finite_CSS.
Proof.
  intros [ ] H3 x Hx.
  split; generalize Hx; clear Hx.
  apply bbi_expansion_CSS    with (2 := H3); auto.
  apply bbi_expansion_finite with (2 := H3); auto.
Qed.

Section bbi_expansion_subst.

  Variable sigma : label -> label.

  Let s a := l (sigma a).

  Section rules_inst.
  
    Variable X : stm -> Prop.
    Variable C : cst -> Prop.

    Hypothesis sigma_inj_1 : forall a b, ~ alphabet C a -> ~ alphabet C b -> sigma a = sigma b -> a = b.
    Hypothesis sigma_inj_3 : pr_image sigma (pd_image sigma (alphabet C)) inc alphabet C.

    Let sigma_inj_2 : forall a, alphabet (subst_dimage s C) (sigma a) -> alphabet C a.
    Proof.
      intros a Ha.
      apply sigma_inj_3.
      exists (sigma a).
      split; auto.
      apply subst_dimage_alphabet, subst_alphabet_prop in Ha.
      destruct Ha as ( b & H1 & [ H2 | [] ] ).
      exists b; auto.
    Qed.

    Let alphabet_lemma a : ~ alphabet C a -> ~ alphabet (subst_dimage s C) (sigma a).
    Proof. intros H; contradict H; auto. Qed.
    
    Hint Resolve subst_pstm_mono.

    Ltac pd_image_tac := 
      match goal with 
        [ |- pd_image _ (_ cup _) ~ _ cup _ ] => try apply id_transitive with (1 := pd_image_id_cup _ _ _ _ _) 
      | [ |- pd_image _ (sg _)    ~ _       ] => try apply id_transitive with (1 := pd_image_sg _ _ _ _)
      | [ _ : pme C ?c |- pd_image ?s _ _ ]   => apply (pd_image_trivial _ _ s _ c)
      end; auto.

    Ltac cup_tac := 
      match goal with 
        [ |- _ cup _ ~ _ cup _ ] => apply bin_union_congr; auto
      end.

    Ltac pme_tac :=
      match goal with
        [ |- pme _ ~ pme _ ] => apply pme_id
      end.

    Ltac unfold_tac :=
      match goal with
        [ |- subst_pstm _ _       ~ _ ] => unfold subst_pstm
      | [ |- (subst_dimage _ _) ~ _ ] => unfold subst_dimage
      end.

    Ltac solve := 
      repeat match goal with 
         [ |- CSS_equiv _ _ ] => split; simpl
       | [ |- CSS_id _ _ ] => split; simpl
       | [ |- _ /\ _ ] => split
       | [ |- _      ] => repeat (unfold_tac || pd_image_tac || cup_tac || pme_tac); auto
      end.
    
    Proposition bbi_exp_subst r p : In r bbi_exp_rules 
                                 -> bbi_fireable r p (X,C) 
                                 -> exists p', bbi_fireable r p' (subst_CSS s (X,C))
                                           /\  listbranch_eq CSS_id
                                                 (map (subst_CSS s) (bbi_expand (X,C) (bbi_inst _ p)))
                                                 (bbi_expand (subst_CSS s (X,C)) (bbi_inst _ p')).
    Proof.
      intros Hr Hp.
      unfold bbi_exp_rules in Hr; In_auto; symmetry in Hr; 
      subst; destruct p as ((((((A,B),a),b),m),x),y); 
      simpl in Hp; auto.
      
      exists (A,B,a,b,subst_w s m,x,y); simpl; solve.
      exists (A,B,a,b,subst_w s m,x,y); simpl; solve.
      exists (A,B,a,b,subst_w s m,x,y); simpl; solve.
      exists (A,B,a,b,subst_w s m,x,y); simpl; solve.
      exists (A,B,a,b,subst_w s m,x,y); simpl; solve.
    
      destruct Hp as ( H1 & H2 & H3 & H4 ).
      exists (A,B,sigma a,sigma b,subst_w s m,x,y); simpl; solve.
      red; solve.
    
      destruct Hp as ( H1 & H2 ).
      exists (A,B,a,b,subst_w s m,subst_w s x,subst_w s y); simpl; solve.
      red; solve.
      apply subst_dimage_pme_inc.
      unfold subst_dimage; rewrite <- subst_o; pd_image_tac.

      destruct Hp as ( H1 & H2 ).
      exists (A,B,a,b,subst_w s m,subst_w s x,subst_w s y); simpl; solve.
      red; solve.
      apply subst_dimage_pme_inc.
      unfold subst_dimage; rewrite <- subst_o; pd_image_tac.

      destruct Hp as ( H1 & H2 & H3 & H4 ).
      exists (A,B,sigma a,sigma b,subst_w s m,x,y); simpl; solve.
      red; solve.
    Qed.

  End rules_inst.

  Section subst_inj.

    Hypothesis sigma_inj : injective sigma.
    
    Let sigma_inj_1 C a b : ~ alphabet C a -> ~ alphabet C b -> sigma a = sigma b -> a = b.
    Proof. intros; apply sigma_inj; auto. Qed.
    
    Let sigma_inj_3 C : pr_image sigma (pd_image sigma (alphabet C)) inc alphabet C.
    Proof.
      intros x (y & (z & H1 & H2) & H3).
      rewrite <- H3 in H2.
      rewrite <- (sigma_inj _ _ H2); auto.
    Qed.

    Theorem bbi_expansion_subst_inj : forall css lCSS, bbi_expansion css lCSS 
                                      -> exists lCSS', bbi_expansion (subst_CSS s css) lCSS' 
                                                    /\ listbranch_eq CSS_id (map (subst_CSS s) lCSS) lCSS'.
    Proof.
      intros (X,C) lCSS [ r p Hr Hp ]. 
      destruct bbi_exp_subst with (C := C) (4 := Hp) as ( p' & H1 & H2 ); auto.
      exists (bbi_expand (subst_CSS s (X,C)) (bbi_inst r p')).
      split; auto.
      constructor; auto.
    Qed.

  End subst_inj.

End bbi_expansion_subst.

Section subst_stable.

  Variable sigma : label -> label.
  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Hypothesis sigma_inj_1 : forall a b, ~ alphabet C a -> ~ alphabet C b -> sigma a = sigma b -> a = b.
  Hypothesis sigma_inj_2 : pr_image sigma (pd_image sigma (alphabet C)) inc alphabet C.

  Let s x := l (sigma x).

  Proposition bbi_expansion_subst_stable lcss : bbi_expansion (X,C) lcss 
                                  -> exists l', bbi_expansion (subst_CSS s (X,C)) l' 
                                             /\ listbranch_eq CSS_id (map (subst_CSS s) lcss) l'.
  Proof.
    intros [ r p Hr Hp ].
    destruct bbi_exp_subst with (sigma := sigma) (C := C) (4 := Hp) as ( p' & H1 & H2 ); auto.

    exists (bbi_expand (subst_CSS s (X,C)) (bbi_inst r p')).
    split; auto; constructor; auto.
  Qed.

End subst_stable.

Section bbi_expansion_cup.

  Variables X Xe : stm -> Prop.
  Variables C Ce : cst -> Prop.

  Let alphabet_lemma a : (alphabet Ce a -> alphabet C a) -> ~ alphabet C a -> ~ alphabet (Ce cup C) a.
  Proof.
    intros H1 H2 H3. apply H2.
    apply alphabet_cup in H3.
    apply bin_union_sg in H3.
    destruct H3; auto.
  Qed.

  Ltac solve_alpha H :=       
    apply alphabet_lemma; auto;
    intro; apply H; pred_solve; 
    apply alphabet_cup, bin_union_inc_r1, alphabet_sg; pred_solve.

  Ltac solve := repeat
       match goal with 
            [ _: ?X ?y          |- (_ cup ?X) ?y  ] => apply bin_union_inc_r2; auto 
          | [ H: _ /\ _         |- _              ] => destruct H
          | [                   |- _ /\ _         ] => split; auto
          | [ H: pme _ ?x       |- pme _ ?x       ] => apply pme_mono with (2 := H); auto
          | [ H: _ cap _ inc _  |- ~ alphabet _ _ ] => solve_alpha H
        end.

  Proposition bbi_expansion_cup_fire r p : In r bbi_exp_rules 
                                        -> alphabet (bbi_tab_constraints (bbi_inst _ p)) cap alphabet Ce inc alphabet C 
                                        -> bbi_fireable r p (X       , C)
                                        -> bbi_fireable r p (Xe cup X, Ce cup C).
  Proof.
    intros Hr Ha Hp.
    unfold bbi_exp_rules in Hr; In_auto; symmetry in Hr; 
    subst; destruct p as ((((((?,?),?),?),?),?),?); 
    simpl in Ha, Hp; auto; simpl; solve; red in Hp |- *;
    solve.
  Qed.

End bbi_expansion_cup.

Proposition bbi_expansion_cup X Xe C Ce lcss : bbi_expansion (X,C) lcss
                                            -> alphabet (bbi_tab_constraints lcss) cap alphabet Ce inc alphabet C 
                                            -> exists lb, bbi_expansion (Xe cup X, Ce cup C) lb 
                                                       /\ listbranch_eq CSS_id (bbi_expand (Xe,Ce) lcss) lb.
Proof.
  intros [ ? ? Hr Hp ]; simpl; intros G.
  apply bbi_expansion_cup_fire with (Xe := Xe) (Ce := Ce) (1 := Hr) in Hp.
  match goal with 
    [ _: bbi_fireable ?r ?p ?css |- _ ] => exists (bbi_expand css (bbi_inst r p)) 
  end.
  split; auto.
  constructor; auto.
  apply inc_transitive with (2 := G); pred_solve.
Qed.



