(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.

Require Import predicate.
Require Import graph.

(* this is a constructive proof of some variant of unique choice principle 

   uniq_choice (S n,x) <-> x0 (R 0) o ... o (R n) x
 
*)

Section limit.

  Local Set Implicit Arguments.

  Variable   X  : Type.
  Variable   x0 : X.
  Variable   R  : nat -> X*X -> Prop.
  Hypothesis HR : forall n, function_graph (R n).

  Definition uniq_choice c := 
    let fix loop n x := match n with
      | 0   => x = x0
      | S n => exists k, R n (k,x) /\ loop n k
    end in match c with (n,x) => loop n x end.

  Theorem uniq_choice_function : function_graph uniq_choice.
  Proof.
    apply functional_total_graph_function.

    intros n; unfold uniq_choice; induction n; intros x y.
    intros H; rewrite H; auto.
    intros (x' & H1 & H2) (y' & H3 & H4).
    rewrite (IHn _ _ H2 H4) in H1.
    apply (function_graph_functional _ _ _ (HR n) _ _ _ H1 H3).
    
    intros n; unfold uniq_choice; induction n.
    exists x0; auto.
    destruct IHn as [ k Hk ].
    destruct (HR n k) as [ y Hy ].
    exists y; exists k.
    apply gr_dimage_sg' in Hy.
    split; auto.
  Qed.

  Theorem uniq_choice_O : uniq_choice (0,x0).
  Proof. red; auto. Qed.

  Theorem uniq_choice_S : forall n x y, R n (x,y) -> uniq_choice (n,x) -> uniq_choice (S n, y).
  Proof. intros ? x ? ? ?; exists x; split; auto. Qed.

End limit.
